import xmltodict
import json
import os
from dict2graph import Dict2graph
from Configs import getConfig
import logging
from uuid import uuid4
from py2neo import Graph
from DZDutils.neo4j import run_periodic_iterate

# from py2neo.client import TransactionError, BrokenTransactionError
from pms.tools import DTDUtil, get_graph
from pms.import_worker.filter import Filter

from typing import Dict, List
from neobulkmp.cache_backend import CacheInterface
from neobulkmp import WorkerSourcing

config = getConfig()


class PubMedXMLParser:
    # Parent Worker process
    worker: WorkerSourcing = None

    def __init__(
        self,
        parent_worker: WorkerSourcing,
        xml_file: str,
        debug: bool = False,
    ):
        self.articles_parsed_count = 0
        self.articles_failed_parsing_count = 0
        self.articles_in_memory_count = 0
        self.articles_loaded_to_db_count = 0
        self.xml_file = xml_file
        self.dtd = DTDUtil(config.XML_DTD_SOURCE)
        self.delete_pmids: List = []
        self.d2g = Dict2graph()
        self.d2g._debug = debug
        self._set_config()
        self.filter = Filter(config.FILTER_RULES)
        self.worker = parent_worker
        self.graph_connection_params = config.NEO4J

    def run(self, cache: CacheInterface):
        self.cache = cache
        self.graph = Graph(**self.graph_connection_params)

        with open(self.xml_file, "rb") as f:
            xmltodict.parse(
                f,
                item_depth=2,
                item_callback=self._handle_article_callback,
                disable_entities=False,
                force_list=self._force_list,
                postprocessor=self._xmltodict_postprocessor,
                attr_prefix="",
                cdata_key="text",
                dict_constructor=dict,
                strip_whitespace=True,
                cdata_separator="###CD###"
                # force_cdata=True,
            )
        # Flush last articles
        self._load_to_db()

    def _node_adding_callback(self, node):
        if node.__primarylabel__ == "PubMedArticle":
            node["xml_source_file"] = os.path.basename(self.xml_file)
        if node.__primarylabel__ == "Keyword":
            if "text" in node and node["text"] is not None:
                node["text_lower"] = node["text"].lower().strip()
        return node

    def _handle_article_callback(self, path, article):

        try:
            if path == [("PubmedArticleSet", None), ("DeleteCitation", None)]:
                self.delete_pmids.extend(article["PMID"])
            elif path == [("PubmedArticleSet", None), ("PubmedArticle", None)]:
                self.d2g.parse(article, "PubMedArticle", instant_save=False)
                # if self.filter.pass_filter(self.d2g):
                #    self.d2g.save()
                self.d2g.save()
                self.articles_parsed_count += 1
                self.articles_in_memory_count += 1
        except Exception as e:
            message = f"Parsing failure for article in xml {self.xml_file}. Article json: \n-------------\n {json.dumps(article)} \n-------------\n"
            self.worker.log.exception(message)
            # log.exception(message)

            self.articles_failed_parsing_count += 1

        if self.articles_in_memory_count >= config.ARTICLE_FLUSH_INTERVAL:
            self._load_to_db()

        return True

    def _load_to_db(self):
        for nodeSet in self.d2g.nodeSets.values():
            self.worker.store_graphSet(nodeSet)
            # self.worker.cache.store_NodeSet(nodeSet)

        for relSet in self.d2g.relationshipSets.values():
            self.worker.store_graphSet(relSet)
            # self.worker.cache.store_RelSet(relSet)

        self.articles_loaded_to_db_count += self.articles_in_memory_count
        self.articles_in_memory_count = 0
        self.d2g.clear()
        self._mark_to_be_deleted_articles()

    def _mark_to_be_deleted_articles(self):
        run_periodic_iterate(
            graph=get_graph(),
            cypherIterate="UNWIND $pmids as pm_id return pm_id",
            cypherAction="MERGE (n:PubmedArticle{PMID:pm_id}) SET n:_PubmedArticle_DELETED",
            parallel=True,
            params={"pmids": self.delete_pmids},
            _log_status_func=self.worker.log.debug,
        )
        self.delete_pmids = []

    def _xmltodict_postprocessor(self, path, key, value):
        # some texts come with html tags in it (<b>,<i>,<sup>...). see https://dtd.nlm.nih.gov/ncbi/pubmed/doc/out/190101/el-b.html
        # xmltodict can not distinguish between xml and html tags.
        # this is a quick, messy, dirty hack. it can destory the correct sequence of html tags texts and hits the perfomance; see https://github.com/martinblech/xmltodict/issues/63 and https://github.com/martinblech/xmltodict/issues/247
        # todo: fix 247 issue
        if key in [
            "ArticleTitle",
            "Affiliation",
            "CollectionTitle",
            "AbstractText",
            "Keyword",
            "CollectiveName",
            "VolumeTitle",
            "Citation",
            "PublisherName",
            "Title",
        ]:
            """
            hack to fix html tags in text
            {[('b', ['Symptomatic']), ('text', 'Unusual  Multipartite Patella Associated with Quadriceps Fat Pad Edema.')]}
            """
            try:
                if isinstance(value, dict):
                    tag_texts = []
                    pubmed_html_tag = [
                        "b",
                        "i",
                        "sup",
                        "sub",
                        "u",
                        "mml:math",
                        "DispFormula",
                    ]
                    if "text" in value:
                        result_val = value["text"]
                    else:
                        result_val = None
                    if result_val is not None:
                        for propname, item in value.items():
                            if propname in pubmed_html_tag:
                                if not isinstance(item, list):
                                    vals = [item]
                                else:
                                    vals = item
                                for val in vals:
                                    if isinstance(val, dict):
                                        val = json.dumps(val)
                                    if val:
                                        result_val = result_val.replace(
                                            "###CD###", val, 1
                                        )

                    else:
                        for propname, item in value.items():
                            if propname in pubmed_html_tag:

                                if not isinstance(item, list):
                                    vals = [item]
                                else:
                                    vals = item
                                result_val += " " + " ".join(item)
                    return (key, result_val)
            except Exception as e:
                self.worker.log.exception(
                    "----------\nError postprocessing {KEY} with value '{VAL}'".format(
                        KEY=key, VAL=value
                    )
                )
                return (key, "PARSER_ERROR_NO_VALUE")

        elif isinstance(value, dict) and "text" in value:
            value["text"] = (
                value["text"].replace("###CD###\n", "").replace("###CD###", "").strip()
            )
            if not value["text"]:
                del value["text"]
        elif isinstance(value, str):
            value = value.replace("###CD###\n", "").replace("###CD###", "").strip()
        return (key, value)

    def _force_list(self, path, key, value):
        """example:
        path:   [('PubmedArticleSet', None), ('PubmedArticle', None), ('MedlineCitation', OrderedDict([('Status', 'Publisher'), ('Owner', 'NLM')])), ('Article', OrderedDict([('PubModel', 'Print-Electronic')])), ('AuthorList', OrderedDict([('CompleteYN', 'Y')]))]
        key:    Author
        value:  OrderedDict([('@ValidYN', 'Y'), ('LastName', 'Kalla'), ('ForeName', 'Roger'), ('Initials', 'R'), ('AffiliationInfo', OrderedDict([('Affiliation', 'Department of Neurology, University Hospital Bern, Bern. Switzerland.')]))])
        """
        if key in ["Chemical"]:
            return False
        return self.dtd.is_list_element(path, key)

    def _set_config(self):
        # MATCH (n) RETURN DISTINCT(labels(n)) ORDER BY labels(n)
        # lock allowed labels to prevent unknown/new xml objects
        self.d2g.config_list_allowlist_nodes = [
            "Abstract",
            "AbstractText",
            # "AccessionNumber",
            "Affiliation",
            "ArticleId",
            "Author",
            "Identifier",
            "Chemical",
            "ChemicalList",
            "Contribution",
            "DataBank",
            "Date",
            "Gene",
            "GeneSymbolList",
            "GeneralNote",
            "Grant",
            "GrantList",
            "ISSN",
            "Investigator",
            "Journal",
            "JournalIssue",
            "Keyword",
            "KeywordList",
            "Language",
            "MedlineJournalInfo",
            "MeshDescriptor",
            "MeshHeading",
            "MeshHeadingList",
            "MeshQualifier",
            "PersonalNameSubject",
            "PersonalNameSubjectList",
            "PubMedArticle",
            "PublicationType",
            "Reference",
            "ReferenceList",
        ]

        self.d2g.config_dict_blocklist_props = {
            "Author": ["ValidYN"],
            "ArticleId": ["ValidYN"],
            "Date": ["DateType"],
            #  "The @DateType will always be present and will always indicate "Electronic"." we can ignore this -> https://dtd.nlm.nih.gov/ncbi/pubmed/doc/out/190101/el-ArticleDate.html
        }

        self.d2g.config_list_allowlist_collection_hubs = ["None"]
        self.d2g.config_dict_label_override = {
            "ArticleDate": {"Date": {"type": "ArticleDate"}},
            "PubDate": {"Date": {"type": "PubDate"}},
            "PubMedPubDate": {"Date": {"type": "PubMedPubDate"}},
            "DateCompleted": {"Date": {"type": "DateCompleted"}},
            "DateRevised": {"Date": {"type": "DateRevised"}},
            "AffiliationInfo": "Affiliation",
            "DescriptorName": "MeshDescriptor",
            "QualifierName": "MeshQualifier",
            "OtherAbstract": {"Abstract": {"OtherAbstract": True}},
            "OtherID": {"ArticleId": {"OtherId": True}},
            "ELocationID": {"ArticleId": {"ELocationID": True}},
            "AuthorListAuthor": "Author",
            "GeneSymbol": "Gene",
        }

        self.d2g.config_dict_property_name_override = {
            "PubMedArticle": {
                "PMIDtext": "PMID",
                # "PublicationTypetext": "PublicationType",
            },
            "ArticleId": {"text": "ID", "Source": "IdType"},
            "Chemical": {"NameOfSubstance-UI": "UI", "text": "Name"},
            "DataBank": {"DataBankName": "Name"},
            "Gene": {"GeneSymbol": "sid"},
            "ISSN": {"text": "ID"},
            "AccessionNumber": {"AccessionNumber": "ID"},
            "Identifier": {"text": "ID"},
            "AbstractText": {"AbstractText": "text"},
        }
        self.d2g.config_list_deconstruction_limit_nodes = []
        self.d2g.config_dict_primarykey_attr_by_label = {
            "PubMedArticle": "PMID",
            "Identifier": "ID",
            "Keyword": "text",
            "ArticleId": "ID",
            "PublicationType": "UI",
            "MeshDescriptor": "UI",
            "MeshQualifier": "UI",
            "Chemical": "UI",
            # "AccessionNumber": "ID",
            "Gene": "Name",
            "ISSN": "ID",
            "MedlineJournalInfo": "MedlineTA",
            "DataBank": "name",
        }
        self.d2g.config_dict_primarykey_generated_hashed_attrs_by_label = {
            "Date": "AllAttributes",
            "JournalIssue": "AllContent",
            "Journal": "AllAttributes",
            "MeshHeadingList": "AllContent",
            "MeshHeading": "AllContent",
            "AbstractText": ["text"],
            "Abstract": "InnerContent",
            "ArticleIdList": "InnerContent",
            "Author": "AllAttributes",
            "KeywordList": "OuterContent",
            "ReferenceList": "AllContent",
            "Grant": "AllAttributes",
            "Investigator": "AllAttributes",
            "PersonalNameSubject": "AllAttributes",
            "PersonalNameSubjectList": "AllContent",
            "ChemicalList": "AllContent",
            "GrantList": "AllContent",
            "GeneralNote": ["text"],
            "Reference": ["Citation"],
            "Affiliation": "AllAttributes",
            "GeneSymbolList": "InnerContent",
        }
        self.d2g.config_dict_flip_nodes = {"Journal": "JournalIssue"}

        self.d2g.config_list_throw_away_nodes_with_empty_key_attr = [
            "Keyword",
            "ArticleId",
            "DataBank",
            # "AccessionNumber",
            "Chemical",
            "MedlineJournalInfo",
        ]
        self.d2g.config_list_throw_away_from_nodes = [
            "History",
            "Pagination",
            "SupplMeshList",
            "CommentsCorrectionsList",
        ]

        self.d2g.config_dict_node_prop_to_rel_prop = {
            "Keyword": {"MajorTopicYN": ["KEYWORDLIST_HAS_KEYWORD"]},
            "MeshDescriptor": {"MajorTopicYN": ["MESHHEADING_HAS_MESHDESCRIPTOR"]},
            "MeshQualifier": {"MajorTopicYN": ["MESHHEADING_HAS_MESHDESCRIPTOR"]},
            "Date": {"type": ["JOURNALISSUE_HAS_DATE", "PUBMEDARTICLE_HAS_DATE"]},
            "Author": {"EqualContrib": ["CONTRIBUTION_HAS_AUTHOR"]},
        }

        # self.d2g.config_dict_concat_list_attr = {"PubMedArticle": {"Language": ", "}}

        # selectively flatten the json
        self.d2g.config_dict_interfold_json_attr = {
            "PubMedArticle": {
                "MedlineCitation": {"combine_attr_names": False},
                "PubmedData": {"combine_attr_names": False},
                "Article": {"combine_attr_names": False},
                "AuthorList": {"combine_attr_names": True},
                "InvestigatorList": {"combine_attr_names": False},
                "ArticleIdList": {"combine_attr_names": False},
                "DataBankList": {"combine_attr_names": False},
                "PMID": {"combine_attr_names": True},
                "PublicationTypeList": {"combine_attr_names": False},
                # "PublicationType": {"combine_attr_names": True},
            },
            "Reference": {"ArticleIdList": {"combine_attr_names": False}},
            # "DataBank": {"AccessionNumberList": {"combine_attr_names": False}},
            "Chemical": {"NameOfSubstance": {"combine_attr_names": False}},
        }

        self.d2g.config_dict_hubbing = {
            "PubMedArticle": [
                {
                    "hub_member_labels": ["Author", "Affiliation"],
                    "hub_label": "Contribution",
                    "hub_id_from": "edge",
                },
                {
                    "hub_member_labels": ["Investigator", "Affiliation"],
                    "hub_label": "Contribution",
                    "hub_id_from": "edge",
                },
            ]
        }

        self.d2g.config_func_node_post_modifier = self._node_adding_callback

        self.d2g.make_distinct_before_insert = False



# ToDo: Rewrite with https://boolean-parser.readthedocs.io/en/latest/intro.html#id1 or https://pyparsing-docs.readthedocs.io/en/latest/HowToUsePyparsing.html

import re
from Configs import getConfig
import logging
from dict2graph import Dict2graph
import operator
from typing import Any, Callable, Union, Dict, List

config = getConfig()
log = logging.getLogger("FilterLOgger")

# sequence of the operators in this dict is important, as we first parse multi char operators
COMPARISON_OPERATOR_DICT: Dict[str, Callable[[Any, Any], bool]] = {
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    "=": operator.eq,
    "<": operator.lt,
    ">": operator.gt,
    " in ": lambda left, right: left in right,
    " not in ": lambda left, right: left not in right,
}

LOGICAL_OPERATOR_DICT: Dict[str, Callable[[Any, Any], bool]] = {
    " and ": lambda left, right: left and right,
    " or ": lambda left, right: left or right,
}


class Filter(object):
    def __init__(
        self,
        filter_rules: Union["Filter.FilterRuleGroup", "Filter.FilterRule", str] = None,
    ):
        if isinstance(filter_rules, str):
            self.Deserializer.from_string(filter_rules, filter=self)
        else:
            self.filter_rules = filter_rules

    def pass_filter(self, article_parser: Dict2graph):
        self.data = article_parser._current_nodes
        return bool(self.filter_rules)

    class FilterRule(object):
        filter = None

        def __init__(
            self,
            filter: "Filter",
            node_label: str,
            prop_name: Union[str, List[str]],
            op: Callable[[Any, Any], bool],
            prop_value: Union[str, int, List[str], List[int]],
        ):
            self.filter = filter
            if isinstance(op, str):
                operator = COMPARISON_OPERATOR_DICT[op]
            else:
                operator = op
            if operator not in list(COMPARISON_OPERATOR_DICT.values()):
                raise ValueError(
                    f"Can not create FilterRule. Unknown operator. Expected {list(COMPARISON_OPERATOR_DICT.keys())} got '{op}'"
                )
            self.op_func: Callable[[Any, Any], Any] = operator
            self.node_label = node_label

            self.prop_name = [prop_name] if isinstance(prop_name, str) else prop_name
            self.prop_value = (
                [prop_value] if isinstance(prop_value, str) else prop_value
            )

            if len(self.prop_name) != len(self.prop_value):
                raise ValueError(
                    f"'{self}' is corrupt. Expected same amount of properties and values."
                )

        def __repr__(self) -> str:
            op_str = list(COMPARISON_OPERATOR_DICT.keys())[
                list(COMPARISON_OPERATOR_DICT.values()).index(self.op_func)
            ]
            if len(self.prop_name) == 1:
                prop = self.prop_name[0]
                val = (
                    self.prop_value[0]
                    if isinstance(self.prop_value[0], (float, int))
                    else f"'{self.prop_value[0]}'"
                )
            else:
                prop = self.prop_name
                val = [
                    v if isinstance(v, (float, int)) else f"'{v}'"
                    for v in self.prop_value
                ]
            return f"{self.node_label}.{prop} {op_str} {val}"

        def __str__(self) -> str:
            return self.__repr__()

        def __bool__(self):
            # actual evaluation of rule applied to data

            # get all nodes matching the label configured in the FilterRule
            nodes = self.filter.data.get(frozenset([self.node_label]), [])
            log.debug(frozenset([self.node_label]))
            log.debug(nodes)

            # Throw these nodes in the filter one by one
            log.debug(f"Check {len(nodes)}")
            for node in nodes:
                # Atm we only apply an "or" logic to a filter. if one of the nodes is true the filter returns true
                if self._resolve_filter(node):
                    return True
            return False

        def _resolve_filter(self, node):
            for index, prop in enumerate(self.prop_name):
                log.debug(f"Check prop {prop}")
                # Check if one of the nodes with a certain label has the expected props and values
                if prop in node:
                    log.debug(
                        f"  {node[prop]} to {self.prop_value[index]} results in {self.op_func(node[prop], self.prop_value[index])}"
                    )
                    if not self.op_func(node[prop], self.prop_value[index]):
                        return False
                else:
                    return False
            return True

    class FilterRuleGroup(object):
        def __init__(
            self,
            filter_rules: List[
                Union["Filter.FilterRule", "Filter.FilterRuleGroup"]
            ] = None,
            operator: str = None,
            parenthesis_group: bool = False,
        ):
            self.op_func = None
            if operator:
                self.operator(operator)
            if filter_rules is None:
                filter_rules = []
            elif not isinstance(filter_rules, list):
                filter_rules = [filter_rules]
            self.rules = filter_rules
            self.parenthesis_group = parenthesis_group

        def operator(self, op: str):
            oper = None
            if op in [" and ", "and", "&", "&&"]:
                oper = LOGICAL_OPERATOR_DICT[" and "]
            if op in [" or ", "or", "|", "||"]:
                oper = LOGICAL_OPERATOR_DICT[" or "]
            if oper is None:
                raise ValueError(
                    f"Unknown rule chain operator. Expected 'and','or' got '{op}'"
                )
            self.op_func = oper

        def get_operator_str(self):
            if self.op_func:
                return list(LOGICAL_OPERATOR_DICT.keys())[
                    list(LOGICAL_OPERATOR_DICT.values()).index(self.op_func)
                ]
            else:
                return None

        def add_rule(self, rule: "Filter.FilterRule"):
            self.rules.append(rule)

        def __repr__(self) -> str:
            op_func = str(self.get_operator_str())
            s = ""
            if self.parenthesis_group:
                s += "("
            s += f"{op_func.join([str(rule) for rule in self.rules])}"
            if self.parenthesis_group:
                s += ")"
            return s

        def __str__(self) -> str:
            return self.__repr__()

        def __bool__(self):
            # Check for sanity. If error occures here, there is possible a bug in the Filter.Deserializer
            if not self.op_func and len(self.rules) > 1:
                raise ValueError(f"Rule group {self} has no logical operator.")
            if not self.rules:
                raise ValueError("Rule group has no rules.")
            # run filter rule group

            operator = self.get_operator_str()
            log.debug(f'Run {str(operator).upper()}-rules "{self.rules}"')
            if operator == " or ":
                return any(self.rules)
            elif operator == " and ":
                return all(self.rules)
            elif len(self.rules) == 1:
                return bool(self.rules[0])
            else:
                raise ValueError(f"Something is wrong with rule group '{self}'")

    class Deserializer(object):
        def __init__(self, filter: "Filter" = None):
            if filter is None:
                self.filter = Filter()
            elif isinstance(filter, Filter):
                self.filter = filter
            else:
                raise ValueError(f"Expected None or Filter type got {type(filter)}")

        def eat(self, data):
            self.filter_serialized = data

        def run(self):
            paranthesis_groups = self._seperate_parenthesis_groups(
                self.filter_serialized
            )
            logical_groups = self._split_logical_groups(paranthesis_groups)
            implicit_logical_groups = self._resolve_implicit_brackets(logical_groups)
            return self._parse_logical_groups(implicit_logical_groups)

        @classmethod
        def from_string(
            cls, filter_rules_string: str, filter: "Filter" = None
        ) -> "Filter":
            des = cls(filter)
            des.eat(filter_rules_string)
            filter.filter_rules = des.run()
            return filter

        def _seperate_parenthesis_groups(self, filter_rules_string) -> List[str]:
            def _concatenate_list(lst):
                conc_sub_list = []
                conc_str = ""
                for item in lst:
                    if isinstance(item, list):
                        conc_sub_list.append(conc_str)
                        conc_str = ""
                        conc_sub_list.append(_concatenate_list(item))
                    else:
                        conc_str += item
                if conc_str:
                    conc_sub_list.append(conc_str)
                return conc_sub_list

            # Parse and group parenthesis
            rule_groups = []
            current_group = rule_groups
            parent_group = None
            grandparent_group = None
            for c in filter_rules_string:
                if c == "(":
                    new_group = []
                    current_group.append(new_group)
                    grandparent_group = parent_group
                    parent_group = current_group
                    current_group = new_group
                elif c == ")":
                    current_group = parent_group
                    parent_group = grandparent_group
                else:
                    current_group += c
            return _concatenate_list(rule_groups)

        def _split_logical_groups(self, parenthesis_groups: List):
            new_list = []
            # print("parenthesis_groups", parenthesis_groups)
            for group in parenthesis_groups:
                if isinstance(group, list):
                    new_list.append(self._split_logical_groups(group))
                elif any(log_op in group for log_op in LOGICAL_OPERATOR_DICT.keys()):
                    # split group with one or more logical operators (and, or) into multiple groups
                    # "foo==2 and bar==1" -> ["foo==2","and", "bar==1"]
                    new_list.extend(
                        re.split(f"({'|'.join(LOGICAL_OPERATOR_DICT.keys())})", group)
                    )
                else:
                    new_list.append(group)
            new_list.reverse()
            return [x for x in new_list if x]  # hack to remove empty entries

        def _parse_logical_groups(
            self, parenthesis_groups: List, implicit_parenthesis=False
        ):
            current_group = Filter.FilterRuleGroup(
                parenthesis_group=not implicit_parenthesis
            )
            for token in parenthesis_groups:
                if token in list(LOGICAL_OPERATOR_DICT.keys()):
                    if current_group.get_operator_str() is None:
                        current_group.operator(token)
                    elif current_group.get_operator_str() != token:
                        raise ValueError(
                            f"Mixed operators in group '{parenthesis_groups}'. Possible a bug in Filter.Deserializer._resolve_implicit_brackets() or syntax error in the filter."
                        )
                elif isinstance(token, list):
                    # step recoursive into child groups and add to current group
                    current_group.add_rule(
                        self._parse_logical_groups(token, implicit_parenthesis=False)
                    )
                    # child_groups.append(self._parse_logical_groups(group))
                elif isinstance(token, tuple):
                    current_group.add_rule(
                        self._parse_logical_groups(token, implicit_parenthesis=True)
                    )
                else:
                    current_group.add_rule(self._parse_rule(token))
            return current_group

        def _resolve_implicit_brackets(self, parenthesis_groups):
            last_operator = None
            current_group = []
            current_sub_group = None
            for token in parenthesis_groups:
                if token == " and " and last_operator == " or ":
                    # we have a operator change to "and". we need to build a subgroup
                    current_sub_group = [current_group.pop()]
                    current_sub_group.append(token)
                    last_operator = token
                elif token == " or " and last_operator == " and ":
                    # we have a operator change to "or". we need to put past tokens in a subgroup or a close an opened subgroup
                    if current_sub_group is None:
                        current_group = [current_group, token]
                    elif current_sub_group:
                        current_group.append(tuple(current_sub_group))
                        current_group.append(token)
                        current_sub_group = None
                    last_operator = token
                elif token in [" and ", " or "]:
                    if current_sub_group:
                        current_sub_group.append(token)
                    else:
                        current_group.append(token)
                    last_operator = token
                else:
                    if isinstance(token, list):
                        statement = self._resolve_implicit_brackets(token)
                    else:
                        statement = token
                    if current_sub_group is None:
                        current_group.append(statement)
                    else:
                        current_sub_group.append(statement)
            # close/append leftover open subgroups
            if current_sub_group is not None:
                current_group.append(tuple(current_sub_group))
            return current_group

        def _parse_rule(self, rule_str: str) -> "Filter.FilterRule":
            operator_str = None
            left = None
            right = None

            for op in COMPARISON_OPERATOR_DICT.keys():
                if op in rule_str:
                    operator_str = op
                    left, right = rule_str.split(op)
                    break
            if None in (operator_str, left, right):
                raise ValueError(
                    f"Can not parse filter rule expected '<left> <operator> <right>' pattern. Got '{rule_str}'"
                )
            label, prop = left.replace('"', "").replace("'", "").split(".")
            operator = COMPARISON_OPERATOR_DICT[operator_str]
            prop = prop.strip().strip('"')
            label = label.strip()
            right = right.strip().strip('"')

            if prop.startswith("[") and prop.endswith("]"):
                prop = prop.strip("[").strip("]")
                prop = prop.split(",")
                prop = [p.strip().strip('"') for p in prop]
                right = right.strip("[").right("]")
                right = re.split(r',(?=")', right)
                right = [r.strip().strip('"') for r in right]
            return Filter.FilterRule(
                filter=self.filter,
                node_label=label,
                prop_name=prop,
                op=operator,
                prop_value=right,
            )


# "MyNode"."MyProp"=="12" or "MyNode"."MyProp"=="13"
# ("MyNode"."MyProp"=="12" or "MyNode"."MyProp"=="13") and "MyNode2"."MyProp2"=="Hello"


#
# rule = 'A1.B=="1Hello" or (MyNode.MyProp==12 or (MyNode.MyProp==13 and MyOtherNode.prop99="waaaass")) and MyNode2.MyProp2=="Hello"'
# res =  [['MyNode2.MyProp2=="Hello"', ' and ', [['MyOtherNode.prop99="waaaass"', ' and ', 'MyNode.MyProp==13'], ' or ', 'MyNode.MyProp==12']], ' or ', 'A1.B=="1Hello"']
# rule = 'a==23 and (b=="sd" and f==True) or c=="d" and (a==23 and (d>3 or f=="sd"))'
# rule = "True and False or True"  # any([all([True,False]),True])
# rule = "True and False and True"  # all([True,False,True])
# print(json.dumps(Filter.from_string(rule), indent=5,))
# rule = "Keyword in ['Laptop']"
# rule = '"Author".[Forename,LastName]==["Fred","Miller"] and Article.id in [12,34,56]'
# rule = '["Fred","Miller"] in Author.[Forename,LastName]'
# rule = '"Fred" in Author.Forename and "Miller" in Author.LastName'
# rule = '"Fred" in Author.Forename and "Miller" in Author.LastName'

# from boolean_parser import parse


# rule = '(Z0.Y0 == "whateva" and A1.B1 == "1Hello") and C2.A2 <= "2wtf" or D3.F3 > 23'
# print(rule)
# print(Filter.Deserializer.from_string(rule))
# print("")

# res = parse(rule)
# print(rule)
# print(res)
# print("")
# rule = 'A1.B1 == "1Hello" and C2.A2 <= "2wtf" or D3.F3 > 23'
# res = parse(rule)
# print(rule)
# print(res)
# print(rule)
# print(Filter.Deserializer.from_string(rule))
# print(rule)

# rule = '(A1.B=="1Hello" or MyNode.MyProp==12) or (MyNode.MyProp==13 and MyOtherNode.prop99="waaaass") and MyNode2.MyProp2=="Hello"'
# print(rule)
# print(Filter.Deserializer.from_string(rule))
# print("")



from Configs import ConfigBase
import os
import logging
import multiprocessing


class DEFAULT(ConfigBase):
    # https://docs.python.org/3/howto/logging.html#logging-levels
    LOG_LEVEL = logging.INFO
    MANAGER_LOGFILE = "./log/manager.log"
    WORKER_LOGFILE = "./log/worker.log"
    STATISTICS_LOGFILE = "./log/stats.log"
    DEBUG = True

    BASE_LINE_MODE = True

    PUBMED_XML_DOWNLOAD_DIR = "./data/"
    # dict with params for https://py2neo.org/v4/database.html#the-graph
    # NEO4J = {"host": "192.168.178.40"}
    NEO4J = {}
    # e.g. NEO4J = {"host"="localhost",password="mypw", port=7688}

    # REDIS Cache for neobulkmp - params as dict. possible parameters see https://github.com/andymccurdy/redis-py/blob/master/redis/client.py#L750
    REDIS = {"host": "localhost"}

    # PUBMED_SOURCE_DIR - str
    # if set to None PMS will download all xml files from the ncbi server.
    # if set to local directory path, all xml from this path will be taken into account instead of downloading them
    #   "/tmp/"
    #   ["/tmp", "/files"]
    # if set to specific file(s) only this xml file be ingested "/tmp/pubmed20n1015.xml"
    #   "/tmp/pubmed20n1015.xml"
    #   ["/tmp/pubmed20n1015.xml","/tmp/pubmed20n1016.xml"]
    # if set to list of FTP url, only these will be downloaded and ingested
    #   "ftp://ftp.ncbi.nlm.nih.gov:21/pubmed/baseline/pubmed20n1015.xml.gz"
    #   ["ftp://ftp.ncbi.nlm.nih.gov:21/pubmed/baseline/pubmed20n1015.xml.gz", "ftp://ftp.ncbi.nlm.nih.gov:21/pubmed/baseline/pubmed20n1012.xml.gz"]
    """PUBMED_SOURCE = [
        "testdata/debug_cases.xml",
        "testdata/single.xml",
    ]"""
    """
    PUBMED_SOURCE = [
        "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed20n1009.xml.gz",
        "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed20n1010.xml.gz",
        "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed20n1012.xml.gz",
        "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed20n1013.xml.gz",
    ]
    """
    """
    PUBMED_SOURCE = [
        "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed20n1009.xml.gz",
        "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed20n1010.xml.gz",
    ]
    """
    # PUBMED_SOURCE = "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed21n0670.xml.gz"

    # Download latest 5 files
    PUBMED_SOURCE = 5

    # Download All
    # PUBMED_SOURCE = None

    # Tiny test
    # PUBMED_SOURCE = "testdata/single.xml"

    REDOWNLOAD_EXTISTING_PUBMED_XMLS = False
    MAX_WORKER_COUNT = multiprocessing.cpu_count()
    MIN_PARSING_WORKERS_COUNT = 3
    MAX_RETRIES_ON_INSERT_ERRORS = 4

    # ARTICLE_FLUSH_INTERVAL
    # load articles to cache every n articles. A higher value spares the cache in terms of connections/requests per second but uses more memory on the parser side
    ARTICLE_FLUSH_INTERVAL = 500
    # commit every n nodes per graphio nodeset
    GRAPHIO_BATCHSIZE = 100000
    # The Document type definition (dtd) file is used in Pubmedsucker to determine wether an elemnt in the source xml is a list or an single attribute.
    # it can be a local file or a url
    # it will be cached locally if it is an url and is used for multiple worker
    XML_DTD_SOURCE = "http://dtd.nlm.nih.gov/ncbi/pubmed/out/pubmed_190101.dtd"

    # PUBMED_FTP_SERVER_CONFIG - dict
    # see config schema from https://git.connect.dzd-ev.de/dzdtools/pythonmodules/tree/master/FTPDownloader

    @property
    def PUBMED_FTP_SERVER_CONFIG(self):
        # quick and dirty solution for https://git.connect.dzd-ev.de/dzdtools/pythonmodules/-/issues/14
        return {
            "url": "ftp.ncbi.nlm.nih.gov",
            "port": 21,
            "user": "anonymous",
            "pw": "anonymous",
            "base_path": None,
            "dest_path": self.PUBMED_XML_DOWNLOAD_DIR,
            "not_to_save_extentions": ["md5", "txt", "html"],
            "overwrite_existing": False,
            "verify_file": True,
            "verify_throw_error_when_failed": True,
            "verify_fallback_to_size_comparison": False,
        }

    # path for base line xmls ( https://www.nlm.nih.gov/databases/download/pubmed_medline.html ) on the NIH ftp server
    PUBMED_BASE_LINE_PATH = "/pubmed/baseline/"
    # path for udpated xmls ( https://www.nlm.nih.gov/databases/download/pubmed_medline.html ) on the NIH ftp server
    PUBMED_UPDATE_PATH = "/pubmed/updatefiles/"

    EXIT_WHEN_DONE = True

    LOADING_LOG_LABELS = ["_PubMedXmlLoadingLog"]

    FORCE_REMERGING_LOADED_PUBMED_XMLS = True

    # This is a limit to prevent the cache on overusing memory. If you have enough free memory you can increase this value
    # if you are short on memory or even had crashed due to low memory it can help to decrease this value.
    # this can impact perfomance
    MAX_NODES_N_RELS_IN_CACHE_BEFORE_PAUSING_PARSING = 10000000

    # Filter rules applied on each article. if positive whole article with all metadata will not be saved
    # FILTER_RULES = 'Author.ForeName == "Connie" and (Keyword.Keyword == "Spinal diseases" or MeshQualifier.text == "physiology")'
    FILTER_RULES = None


class DEV(DEFAULT):
    MAX_NODES_N_RELS_IN_CACHE_BEFORE_PAUSING_PARSING = 100000
    BASE_LINE_MODE = False



import os
import time
import json
import urllib
from Configs import getConfig
from linetimer import CodeTimer
import logging
from logging import Logger
import sys


from neobulkmp import WorkerSourcing
from pms.import_worker.downloader import download
from pms.import_worker.parser import PubMedXMLParser
from py2neo import Node, Graph

config = getConfig()

logging.basicConfig(
    level=config.LOG_LEVEL,
    format="%(asctime)-15s %(processName)-8s %(module)-8s %(levelname)-8s:  %(message)s",
    handlers=[logging.StreamHandler((sys.stdout))],
)


class Worker(WorkerSourcing):
    log: Logger = None

    def _gather_source_data(self, xml_url: str):

        source_file = None
        # Check if argument is a local file or an url
        if os.path.isfile(xml_url):
            self.log.debug(f"{xml_url} is a local file. No download needed")
            source_file = xml_url

        elif urllib.parse.urlparse(xml_url).scheme in ["ftp", "http", "https", "ftps"]:
            source_file = download(xml_url)
        else:
            raise ValueError(
                f"Invalid source url/file. Expected HTTP/FTP-URL or local Path, got '{xml_url}'"
            )
        # Check if file is existent and accessible
        try:
            f = open(source_file)
        except:
            raise ValueError(f"Can not access '{source_file}'")
        finally:
            f.close()
        self.source_file = source_file

    def run(self):
        self.cache = self.cache_backend(self.cache_backend_params)
        self.log: logging.Logger = self.get_logger()
        self.log.addHandler(logging.FileHandler(config.WORKER_LOGFILE))
        self.graph = Graph(**config.NEO4J)
        self._gather_source_data(self.worker_parameter["xml_url"])
        if (
            self._fetch_loading_log() is not None
            and not config.FORCE_REMERGING_LOADED_PUBMED_XMLS
        ):
            self.log.info(
                "Skip '{}' . File was allready parsed and loaded into DB. ...".format(
                    self.source_file
                )
            )
            return
        self.log.info(f"Start parsing {self.worker_parameter['xml_url']} gogogoooo!")
        self.parser = PubMedXMLParser(parent_worker=self, xml_file=self.source_file)

        self.parser.run(self.cache)
        self._create_loading_log()

    def _create_node(self, labels: "list(str)", props: dict):
        # this is a little helper function to create some log,metadata nodes
        tx = self.graph.begin()
        tx.create(Node(*labels, **props))
        tx.commit()

    def _create_loading_log(self):
        self._create_node(
            config.LOADING_LOG_LABELS,
            {
                "filename": str(os.path.basename(self.source_file)),
                "timestamp": time.strftime("%Y-%m-%d %H:%M:%S %z"),
                "PubMedParser_Version": config.VERSION,
                "article_deletions_count": len(self.parser.delete_pmids),
                "to_be_deleted_articles": self.parser.delete_pmids,
            },
        )

    def _fetch_loading_log(self):
        query = "MATCH (n:{}) WHERE n.filename='{}' return n".format(
            ":".join(config.LOADING_LOG_LABELS),
            str(os.path.basename(self.source_file)),
        )
        tx = self.graph
        cursor = tx.run(query)
        result = cursor.data()
        if len(result) == 0:
            return None
        else:
            return result[0]



import os
from os.path import expanduser
import shutil
import logging
import urllib.request as request
from contextlib import closing
from Configs import getConfig


config = getConfig()

log = logging.getLogger("PubmedSuckerETLProcessToFile")


def _extract_gz_file(source_file, target_file):
    import gzip
    import shutil

    # Check sanity
    if not os.path.isfile(source_file):
        log.error(f"Error extracting {source_file}. File does not exists.")
        raise ValueError(f"Error extracting {source_file}. File does not exists.")
    if not os.access(os.path.dirname(target_file), os.W_OK):
        log.error(
            f"Error extracting {source_file}. Target dir {os.path.dirname(target_file)} is not writable."
        )
        raise ValueError(
            f"Error extracting {source_file}. Target dir {os.path.dirname(target_file)} is not writable."
        )
    # try writing file
    try:
        f = open(target_file, "w")
        f.write("test")
        f.close()
        os.remove(target_file)
    except:
        log.error(
            f"Error extracting {source_file}. Target {target_file} is not writable."
        )
        raise ValueError(
            f"Error extracting {source_file}. Target {target_file} is not writable."
        )

    try:
        with gzip.open(source_file, "rb") as f_in:
            with open(target_file, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
    except:
        log.error(f"Error extracting {source_file} to {target_file}")
        raise
    return target_file


def download(source) -> str:
    storage_dir = os.path.abspath(config.PUBMED_XML_DOWNLOAD_DIR)
    os.makedirs(storage_dir, exist_ok=True)
    source_filename = os.path.basename(source)
    download_path = os.path.join(storage_dir, source_filename)
    file_name, file_extension = os.path.splitext(source_filename)
    if file_extension == ".gz":
        storage_path = os.path.join(storage_dir, file_name)
    else:
        storage_path = download_path
    if not os.path.isfile(storage_path) or config.REDOWNLOAD_EXTISTING_PUBMED_XMLS:
        try:
            with closing(request.urlopen(source)) as r:
                with open(download_path, "wb") as f:
                    shutil.copyfileobj(r, f)
        except:
            log.error(f"Failed downloading file from '{source}' to '{download_path}'")
            raise
        if file_extension == ".gz":
            _extract_gz_file(download_path, storage_path)
            os.remove(download_path)
    return storage_path



"""PubMedsucker - PMS

https://git.connect.dzd-ev.de/dzdtools/pubmedsucker

Copyright 2020 Deutsches Zentrum für Diabetesforschung e.V.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


# TODO
# * handle PubmedBookArticleSet


import os
import sys
import logging
import graphio
import py2neo
from linetimer import CodeTimer

from Configs import getConfig


def run():
    from pms.tools import create_config_dirs, get_phase1_file_list, get_graph
    import pms
    from pms.config import DEFAULT

    config: DEFAULT = getConfig()
    create_config_dirs()

    from pms.final_transforming import delete_old_articles, run_transforming_queries
    from pms.import_worker.worker import Worker
    import neobulkmp
    from getversion import get_module_version

    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)-15s %(processName)-8s %(module)-8s %(levelname)-8s:  %(message)s",
        handlers=[logging.StreamHandler((sys.stdout))],
    )

    log = logging.getLogger("PubmedSuckerManagerProcess")
    log.addHandler(logging.FileHandler(config.MANAGER_LOGFILE, mode="a"))

    statistic_logger = logging.getLogger("PMSStatistics")
    statistic_logger.addHandler(logging.FileHandler(config.STATISTICS_LOGFILE))

    log.info(f"Pubmedsucker version: '{get_module_version(pms)[0]}'")
    log.info(f"\t neobulkmp version: '{get_module_version(neobulkmp)[0]}'")
    log.info(f"\t py2neo version: '{get_module_version(py2neo)[0]}'")
    log.info(f"\t graphio version: '{get_module_version(graphio)[0]}'")

    log.debug("PUBMED_SOURCE: {}".format(config.PUBMED_SOURCE))
    try:
        article_count_before = (
            get_graph()
            .begin()
            .run("MATCH (a:PubMedArticle) return count(a) as cnt")
            .data()[0]["cnt"]
        )
    except Exception as e:
        log.error(
            "Error: Can not reach neo4j DB '{}'. Aborting import...".format(
                config.NEO4J
            )
        )
        raise
    article_deleted_count = 0
    node_deleted_count = 0
    timer = CodeTimer("CompleteProcessTimer", unit="s", silent=True)

    with timer:
        p1timer = CodeTimer("Phase1ProcessTimer", unit="s", silent=True)
        with p1timer:
            # Start PHASE 1 - Loading Phase
            log.info("#--------START LOADING PHASE  (1/2)---------#")
            worker_manager = neobulkmp.Manager(
                worker_sourcing_class=Worker,
                worker_parameters=get_phase1_file_list(),
                cache_backend_params=config.REDIS,
                debug=config.DEBUG,
            )
            try:
                worker_manager.cache.test_connection()
            except Exception as e:
                log.error(f"Redis not available:\n {type(e).__name__}: {str(e)}")
                exit(1)
            worker_manager.strategy.max_processes_count = config.MAX_WORKER_COUNT
            worker_manager.strategy.max_graphobjects_count_in_cache = (
                config.MAX_NODES_N_RELS_IN_CACHE_BEFORE_PAUSING_PARSING
            )
            worker_manager.worker_loading_class.amount_of_retries_on_loading_fail = (
                config.MAX_RETRIES_ON_INSERT_ERRORS
            )
            worker_manager.strategy._last_amount_sourcing_cores = (
                config.MIN_PARSING_WORKERS_COUNT
            )
            worker_manager.report_statistics_print_func = statistic_logger.info
            worker_manager.strategy.graphio_batchsize = config.GRAPHIO_BATCHSIZE
            worker_manager.worker_log_level = config.LOG_LEVEL
            worker_manager.create_indexes = True
            worker_manager.merge(graph_params=config.NEO4J)
            # END OF PHASE 1

            # Start PHASE 2 - polishing transforming, delete old data and artefacts
        p2timer = CodeTimer("Phase3ProcessTimer", unit="s", silent=True)
        with p2timer:
            log.info("#--------START TRANSFORMING PHASE (2/2)---------#")
            delete_old_articles()
            run_transforming_queries()

    article_count_after = (
        get_graph()
        .begin()
        .run("MATCH (a:PubMedArticle) return count(a) as cnt")
        .data()[0]["cnt"]
    )

    log.info(
        "Inserted {} Articles in {} Minutes and {} Seconds. Article count in total: {}. Deleted {} Articles({} Nodes).".format(
            article_count_after - article_count_before,
            int(timer.took / 60),
            int(timer.took % 60),
            article_count_after,
            article_deleted_count,
            node_deleted_count,
        )
    )

    log.info(
        " Phase 1 loading took {} Minutes and {} Seconds. thats {}% of the total time".format(
            int(p1timer.took / 60),
            int(p1timer.took % 60),
            round((p1timer.took / timer.took) * 100, 4),
        )
    )

    log.info(
        " Phase 2 loading took {} Minutes and {} Seconds. thats {}% of the total time".format(
            int(p2timer.took / 60),
            int(p2timer.took % 60),
            round((p2timer.took / timer.took) * 100, 4),
        )
    )

    # rancher forced autorestart dirty quickfix
    if not config.EXIT_WHEN_DONE:
        import time

        while True:
            time.sleep(60)


if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    SCRIPT_DIR = os.path.join(SCRIPT_DIR, "..")
    sys.path.append(os.path.normpath(SCRIPT_DIR))
    run()



from setuptools import setup
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="PubMedSucker",
    description="Pull all articles from PubMed and insert them into a Neo4j Graph Database",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.connect.dzd-ev.de/dzdtools/pubmedsucker",
    author="TB",
    author_email="tim.bleimehl@helmholtz-muenchen.de",
    license="MIT",
    packages=["pms"],
    install_requires=[
        "py2neo",
        "lxml",
        "xmltodict",
        "memoization",
        "graphio",
        "neobulkmp",
        "getversion",
        "dict2graph",
        "DZDConfigs",
        "FTPDownloader",
        "DZDutils",
    ],
    python_requires=">=3.9",
    zip_safe=False,
    include_package_data=True,
    use_scm_version={
        "root": ".",
        "relative_to": __file__,
        # "local_scheme": "node-and-timestamp"
        "local_scheme": "no-local-version",
        "write_to": "version.py",
    },
    setup_requires=["setuptools_scm"],
)



import os


from Configs import getConfig
from py2neo import Graph
from FTPDownloader import FTPDownloader

## DTDUtils imports
import io
from lxml import etree
from memoization import cached
from time import sleep

config = getConfig()


def create_config_dirs():
    pathes = [
        config.MANAGER_LOGFILE,
        config.WORKER_LOGFILE,
        config.STATISTICS_LOGFILE,
    ]
    for path in pathes:
        if "." in os.path.basename(path):
            # TODO: hats a dirty,dirty hack! but works for now
            path = os.path.dirname(path)
        os.makedirs(path, exist_ok=True)


def get_graph():
    return Graph(**config.NEO4J)


def get_phase1_file_list():
    file_list = []
    ftp_conf = config.PUBMED_FTP_SERVER_CONFIG
    if config.BASE_LINE_MODE:
        ftp_conf["base_path"] = config.PUBMED_BASE_LINE_PATH
    else:
        ftp_conf["base_path"] = config.PUBMED_UPDATE_PATH
    if isinstance(config.PUBMED_SOURCE, list):
        # list of local or remore files
        for source in config.PUBMED_SOURCE:
            file_list.append({"xml_url": source})
    elif isinstance(config.PUBMED_SOURCE, str):
        if os.path.isdir(config.PUBMED_SOURCE):
            # directory of local files
            for file in [
                f
                for f in os.listdir(config.PUBMED_SOURCE)
                if os.path.isfile(os.path.join(config.PUBMED_SOURCE, f))
                and f.endswith("xml")
            ]:
                file_list.append({"xml_url": os.path.join(config.PUBMED_SOURCE, file)})
        else:
            # single remote or lcoal file
            file_list.append({"xml_url": config.PUBMED_SOURCE})
    elif isinstance(config.PUBMED_SOURCE, int) or config.PUBMED_SOURCE is None:
        # first n files of remote dir
        downloader = FTPDownloader.from_config_dict(ftp_conf)
        remote_file_list = downloader.get_remote_file_list(include_full_url=True)
        if config.PUBMED_SOURCE is None:
            file_list = [{"xml_url": url} for url in remote_file_list]
        else:
            file_list = [
                {"xml_url": url}
                for url in remote_file_list[config.PUBMED_SOURCE * -1 :]
            ]
    return file_list


class DTDUtil(object):
    cache_path = ".cache/dtd_util"
    no_cache = False

    def __init__(self, dtd_file):
        try:
            self._source_file = self.download_dtd(dtd_file)
        except ValueError:  # invalid URL. use path as local file
            self._source_file = dtd_file
        self._parse_dtd()

    def download_dtd(self, url: str):
        import urllib.request

        # TODO: urllib.request.urlopen can be very slow if ipv6 is enabled. investigate workaround
        url_parsed = urllib.parse.urlparse(url)
        filename = os.path.basename(url_parsed.path)
        cache_dir = os.path.join(self.cache_path, url_parsed.netloc)
        cache_path = os.path.join(cache_dir, filename)
        if os.path.isfile(cache_path):
            # file is allready downloaded
            # bad hack. wait some ms if file is still in downloading request by another process

            # tb: 2020_11_26...this still make problems on high scaling (many workers). hotfixed the hack by preloading dtd in the main.py
            sleep(1)
            return cache_path
        os.makedirs(cache_dir, exist_ok=True)
        with urllib.request.urlopen(url) as response:
            with open(cache_path, "w") as target_file:
                target_file.write(response.read().decode("utf-8"))
        # content = response.read().decode("utf-8")
        return cache_path

    def _parse_dtd(self):
        if isinstance(self._source_file, str):
            try:
                self.dtd = etree.DTD(self._source_file)
            except:
                raise
        else:
            # TODO: I dont why i need to reconvert to a string->StringIO ("io.StringIO(file.getvalue())") object here when running in a sub processes? It does not work when directly using the "file"-io.stringIO object (Which is not empty or so).
            # Investigate!
            self.dtd = etree.DTD(io.StringIO(self._source_file.getvalue()))

    @cached
    def is_list_element(self, path, key):

        """example vals
        PATH [('PubmedArticleSet', None), ('PubmedArticle', None), ('MedlineCitation', OrderedDict([('Status', 'Publisher'), ('Owner', 'NLM')])), ('Article', OrderedDict([('PubModel', 'Print-Electronic')])), ('AuthorList', OrderedDict([('CompleteYN', 'Y')]))]
        KEY Author
        """
        if len(path) <= 1:
            # we are at the root element. it can not be a list element
            return False

        parent_key = path[len(path) - 1][0]
        dtd_el = next(
            (x for x in self.dtd.iterelements() if x.name == key),
            None,
        )
        if dtd_el is None:
            return False
        parent_dtd_el = next(
            (x for x in self.dtd.iterelements() if x.name == parent_key),
            None,
        )
        # get a binary tree of contents types of parent
        # BTW: Fuck binary trees in python
        # http://lxml.de/validation.html#id1

        """
        <!ELEMENT	Article (Journal,ArticleTitle,((Pagination, ELocationID*) | ELocationID+),
                     Abstract?,AuthorList?, Language+, DataBankList?, GrantList?,
                     PublicationTypeList, VernacularTitle?, ArticleDate*) >
        """
        """<!ELEMENT	BookDocument ( PMID, ArticleIdList, Book, LocationLabel*, ArticleTitle?, VernacularTitle?,
                        Pagination?, Language*, AuthorList*, InvestigatorList?, PublicationType*, Abstract?, Sections?, KeywordList*, 
                        ContributionDate?, DateRevised?, GrantList?, ItemList*, ReferenceList*) >
        """
        """<!ELEMENT	PubmedArticleSet ((PubmedArticle | PubmedBookArticle)+, DeleteCitation?) >
            <!ATTLIST       PubmedArticleSet
            >

            <!ELEMENT	PubmedArticleSet ((PubmedArticle | PubmedBookArticle)+, DeleteCitation?) >
            <!ATTLIST       PubmedArticleSet
            >
        """

        dtd_el, is_multi = next(
            (
                x
                for x in self.traverse_content_elements(parent_dtd_el.content, False)
                if x[0].name == key
            ),
            (dtd_el, None),
        )
        return is_multi or False

    def traverse_content_elements(self, tree, is_multi):

        if tree.occur in ["mult", "plus"]:
            is_multi = True
        if tree.left is not None:
            yield from self.traverse_content_elements(tree.left, is_multi)

        yield tree, is_multi

        if tree.right is not None:
            yield from self.traverse_content_elements(tree.right, is_multi)



import logging
from pms.tools import get_graph
from Configs import getConfig
from DZDutils.neo4j import run_periodic_iterate

config = getConfig()

log = logging.getLogger("PubmedSuckerFinalTransform")


def delete_old_articles() -> int:
    graph = get_graph()
    del_query = """UNWIND $pmids as pmid 
                MATCH p=(a:_PubmedArticle_DELETED)-[*1..2]->(n) 
                WHERE n:ELocationID OR n:Abstract OR n:AbstractText OR n:ArticleId OR n:ReferenceList OR n:Reference OR n:GrantList OR n:GeneralNote OR n:Contribution OR n:KeywordList OR n:MeshHeadingList OR n:MeshHeading OR n:PersonalNameSubjectList
                DETACH DELETE p
                return count(p) as cnt"""
    log.info(f"RUN 'delete'-QUERY:\n'{del_query}'")
    return graph.run(del_query).data()[0]["cnt"]


def run_transforming_queries():
    log.info("Detach MedLineJournal info from article and connect it to Journal")
    run_periodic_iterate(
        graph=get_graph(),
        cypherIterate="MATCH (p:PubMedArticle)-[r:PUBMEDARTICLE_HAS_JOURNALISSUE]->(ji:JournalIssue)-[jij:JOURNALISSUE_HAS_JOURNAL]->(j:Journal) return p as article,j as journal",
        cypherAction="MATCH (article)-[rj:PUBMEDARTICLE_HAS_MEDLINEJOURNALINFO]->(mji:MedlineJournalInfo) MERGE (journal)-[:JOURNAL_HAS_MEDLINEJOURNALINFO]->(mji) DELETE rj",
        batchSize=100,
        parallel=False,
    )



from DZDutils.neo4j import run_periodic_iterate
import py2neo

delete_pmids = []

run_periodic_iterate(
    graph=py2neo.Graph(),
    cypherIterate="UNWIND $pmids as pm_id return pm_id",
    cypherAction="MERGE (n:PubmedArticle{PMID:pm_id}) SET n:_PubmedArticle_DELETED",
    parallel=True,
    params={"pmids": delete_pmids},
)
